"""Tests for the safir.logging module."""

from __future__ import annotations

import json
import logging
import re
from datetime import UTC, datetime, timedelta
from pathlib import Path
from unittest.mock import ANY

import pytest
import structlog
from _pytest.capture import CaptureFixture
from _pytest.logging import LogCaptureFixture
from httpx import AsyncClient
from pydantic import BaseModel, ValidationError

from safir import logging as safir_logging
from safir.logging import LogLevel, Profile, configure_logging
from safir.testing.uvicorn import spawn_uvicorn

# Small app used to test Uvicorn logging configuration.
_UVICORN_APP = """
from fastapi import FastAPI
from safir.logging import configure_uvicorn_logging


app = FastAPI()
configure_uvicorn_logging()


@app.get("/")
async def root():
    return {"message": "Hello World"}


@app.get("/fail")
async def fail():
    raise TypeError("Uncaught exception")
"""


def _strip_color(string: str) -> str:
    """Strip ANSI color escape sequences.

    See https://stackoverflow.com/questions/14693701/
    """
    return re.sub(r"(?:\x1B[@-_]|[\x80-\x9F])[0-?]*[ -/]*[@-~]", "", string)


def test_log_level_enum() -> None:
    """Test that any case is allowed when initializing the enum."""
    assert LogLevel("warning") == LogLevel.WARNING
    assert LogLevel("WARNING") == LogLevel.WARNING
    assert LogLevel("Error") == LogLevel.ERROR

    # Check that this also works when going through Pydantic.
    class Model(BaseModel):
        log_level: LogLevel

    model = Model.model_validate({"log_level": "warning"})
    assert model.log_level == LogLevel.WARNING
    model = Model.model_validate({"log_level": "inFO"})
    assert model.log_level == LogLevel.INFO

    with pytest.raises(ValidationError):
        Model.model_validate({"log_level": "unknown"})


def test_configure_logging_development(caplog: LogCaptureFixture) -> None:
    """Test that development-mode logging is key-value formatted."""
    caplog.set_level(logging.INFO)

    configure_logging(
        name="myapp", profile=Profile.development, log_level=LogLevel.INFO
    )
    assert safir_logging.logger_name == "myapp"

    logger = structlog.get_logger("myapp")
    logger = logger.bind(answer=42)
    logger.info("Hello world")

    app, level, line = caplog.record_tuples[0]
    assert app == "myapp"
    assert level == logging.INFO
    expected = "[info     ] Hello world                    [myapp] answer=42"
    assert _strip_color(line) == expected


def test_configure_logging_dev_timestamp(caplog: LogCaptureFixture) -> None:
    """Test development-mode logging with an added timestamp."""
    caplog.set_level(logging.INFO)

    configure_logging(
        name="myapp",
        profile="development",
        log_level="info",
        add_timestamp=True,
    )

    logger = structlog.get_logger("myapp")
    logger = logger.bind(answer=42)
    logger.info("Hello world")

    assert caplog.record_tuples[0][0] == "myapp"
    assert caplog.record_tuples[0][1] == logging.INFO
    match = re.match(
        (
            r"(\d+-\d+-\d+T\d+:\d+:[\d.]+Z) \[info\s+\] Hello world \s+"
            r" \[myapp\] answer=42"
        ),
        _strip_color(caplog.record_tuples[0][2]),
    )
    assert match
    isotimestamp = match.group(1)
    assert isotimestamp.endswith("Z")
    timestamp = datetime.fromisoformat(isotimestamp[:-1])
    timestamp = timestamp.replace(tzinfo=UTC)
    now = datetime.now(tz=UTC)
    assert now - timedelta(seconds=5) < timestamp < now


def test_configure_logging_production(caplog: LogCaptureFixture) -> None:
    """Test that production-mode logging is JSON formatted."""
    caplog.set_level(logging.INFO)

    configure_logging(
        name="myapp", profile=Profile.production, log_level=LogLevel.INFO
    )

    logger = structlog.get_logger("myapp")
    logger = logger.bind(answer=42)
    logger.info("Hello world")

    assert caplog.record_tuples[0] == (
        "myapp",
        logging.INFO,
        '{"answer": 42, "event": "Hello world", "logger": "myapp", '
        '"severity": "info"}',
    )


def test_configure_logging_prod_timestamp(caplog: LogCaptureFixture) -> None:
    """Test development-mode logging with an added timestamp."""
    caplog.set_level(logging.INFO)

    configure_logging(
        name="myapp",
        profile="production",
        log_level="info",
        add_timestamp=True,
    )

    logger = structlog.get_logger("myapp")
    logger = logger.bind(answer=42)
    logger.info("Hello world")

    assert caplog.record_tuples[0][0] == "myapp"
    assert caplog.record_tuples[0][1] == logging.INFO
    data = json.loads(caplog.record_tuples[0][2])
    assert data == {
        "answer": 42,
        "event": "Hello world",
        "logger": "myapp",
        "severity": "info",
        "timestamp": ANY,
    }
    assert data["timestamp"].endswith("Z")
    timestamp = datetime.fromisoformat(data["timestamp"][:-1])
    timestamp = timestamp.replace(tzinfo=UTC)
    now = datetime.now(tz=UTC)
    assert now - timedelta(seconds=5) < timestamp < now


def test_configure_logging_level(caplog: LogCaptureFixture) -> None:
    """Test that the logging level is set."""
    caplog.set_level(logging.DEBUG)

    configure_logging(name="myapp", log_level=LogLevel.INFO)
    logger = structlog.get_logger("myapp")

    logger.info("INFO message")
    assert len(caplog.record_tuples) == 1

    # debug-level shouldn't get logged
    logger.debug("DEBUG message")
    assert len(caplog.record_tuples) == 1


def test_duplicate_handlers(capsys: CaptureFixture[str]) -> None:
    """Test that configuring logging more than once doesn't duplicate logs."""
    configure_logging(name="myapp", profile="production", log_level="info")
    configure_logging(name="myapp", profile="production", log_level="info")

    logger = structlog.get_logger("myapp")

    logger.info("INFO not duplicate message")
    captured = capsys.readouterr()
    assert len(captured.out.splitlines()) == 1


def test_dev_exception_logging(caplog: LogCaptureFixture) -> None:
    """Test that exceptions are properly logged in the development logger."""
    configure_logging(
        name="myapp", profile=Profile.development, log_level=LogLevel.INFO
    )
    logger = structlog.get_logger("myapp")

    try:
        raise ValueError("this is some exception")
    except Exception:
        logger.exception("exception happened", foo="bar")

    assert caplog.record_tuples[0][0] == "myapp"
    assert caplog.record_tuples[0][1] == logging.ERROR
    assert "Traceback (most recent call last)" in caplog.record_tuples[0][2]
    assert '"this is some exception"' in caplog.record_tuples[0][2]


def test_production_exception_logging(caplog: LogCaptureFixture) -> None:
    """Test that exceptions are properly logged in the production logger."""
    configure_logging(
        name="myapp", profile=Profile.production, log_level=LogLevel.INFO
    )
    logger = structlog.get_logger("myapp")

    try:
        raise ValueError("this is some exception")
    except Exception:
        logger.exception("exception happened", foo="bar")

    assert caplog.record_tuples[0][0] == "myapp"
    assert caplog.record_tuples[0][1] == logging.ERROR
    data = json.loads(caplog.record_tuples[0][2])
    assert data == {
        "event": "exception happened",
        "exception": ANY,
        "foo": "bar",
        "logger": "myapp",
        "severity": "error",
    }
    assert "Traceback (most recent call last)" in data["exception"]
    assert '"this is some exception"' in data["exception"]


@pytest.mark.asyncio
async def test_uvicorn_logging(tmp_path: Path) -> None:
    """Test the Uvicorn logging configuration."""
    app_path = tmp_path / "test.py"
    app_path.write_text(_UVICORN_APP)
    uvicorn = spawn_uvicorn(
        working_directory=tmp_path, app="test:app", capture=True
    )

    # Perform some operations against the app. Assume the pipe buffers are
    # large enough to hold all of the Uvicorn logging output generated by
    # this. (If they're ever not, we'll have to call communicate on the Popen
    # object.)
    try:
        async with AsyncClient() as client:
            r = await client.get(f"{uvicorn.url}/")
            assert r.status_code == 200
            r = await client.get(f"{uvicorn.url}/fail")
            assert r.status_code == 500
    finally:
        uvicorn.process.terminate()
        stdout, stderr = uvicorn.process.communicate()

    # stdout will be the access log. The raw event will contain the client
    # port, to which we do not have easy access, so only examine the
    # additional fields that should be added by Safir's custom Uvicorn log
    # handling.
    messages = [json.loads(line) for line in stdout.strip().split("\n")]
    assert messages == [
        {
            "event": ANY,
            "severity": "info",
            "httpRequest": {
                "protocol": "HTTP/1.1",
                "requestMethod": "GET",
                "requestUrl": "/",
                "status": "200",
            },
        },
        {
            "event": ANY,
            "severity": "info",
            "httpRequest": {
                "protocol": "HTTP/1.1",
                "requestMethod": "GET",
                "requestUrl": "/fail",
                "status": "500",
            },
        },
    ]

    # stderr will be the error log. There will be a bunch of random Uvicorn
    # messages here; just make sure they're all valid JSON. But make sure that
    # the log for the exception we raised includes the full traceback.
    messages = [json.loads(line) for line in stderr.strip().split("\n")]
    for message in messages:
        if "Exception in ASGI application" in message["event"]:
            assert "TypeError" in message["exception"]
            assert "in fail" in message["exception"]



"""Tests for the database session dependency."""

from __future__ import annotations

import os
from typing import Annotated

import pytest
import structlog
from fastapi import Depends, FastAPI
from httpx import ASGITransport, AsyncClient
from sqlalchemy import Column, String
from sqlalchemy.ext.asyncio import async_scoped_session
from sqlalchemy.future import select
from sqlalchemy.orm import declarative_base

from safir.database import (
    create_async_session,
    create_database_engine,
    initialize_database,
)
from safir.dependencies.db_session import db_session_dependency

TEST_DATABASE_URL = os.environ["TEST_DATABASE_URL"]
TEST_DATABASE_PASSWORD = os.getenv("TEST_DATABASE_PASSWORD")

Base = declarative_base()


class User(Base):
    """Tiny database table for testing."""

    __tablename__ = "user"

    username: str = Column(String(64), primary_key=True)


@pytest.mark.asyncio
async def test_session() -> None:
    logger = structlog.get_logger(__name__)
    engine = create_database_engine(TEST_DATABASE_URL, TEST_DATABASE_PASSWORD)
    await initialize_database(engine, logger, schema=Base.metadata, reset=True)
    session = await create_async_session(engine, logger)
    await db_session_dependency.initialize(
        TEST_DATABASE_URL, TEST_DATABASE_PASSWORD
    )

    app = FastAPI()

    @app.post("/add")
    async def add(
        session: Annotated[
            async_scoped_session, Depends(db_session_dependency)
        ],
    ) -> None:
        async with session.begin():
            session.add(User(username="foo"))

    @app.get("/list")
    async def get_list(
        session: Annotated[
            async_scoped_session, Depends(db_session_dependency)
        ],
    ) -> list[str]:
        async with session.begin():
            result = await session.scalars(select(User.username))
            return list(result.all())

    transport = ASGITransport(app=app)  # type: ignore[arg-type]
    base_url = "https://example.com"
    async with AsyncClient(transport=transport, base_url=base_url) as client:
        r = await client.get("/list")
        assert r.status_code == 200
        assert r.json() == []

        r = await client.post("/add")
        assert r.status_code == 200

        r = await client.get("/list")
        assert r.status_code == 200
        assert r.json() == ["foo"]

    # Retrieve the database contents through an entirely separate connection
    # pool to check the transaction was really committed.
    async with session.begin():
        result = await session.scalars(select(User.username))
        assert result.all() == ["foo"]

    await session.remove()
    await engine.dispose()
    await db_session_dependency.aclose()



"""Test fixtures."""

from __future__ import annotations

from collections.abc import AsyncIterator, Iterator
from datetime import timedelta

import pytest
import pytest_asyncio
import redis.asyncio as redis
import respx

from safir.testing.gcs import MockStorageClient, patch_google_storage
from safir.testing.kubernetes import MockKubernetesApi, patch_kubernetes
from safir.testing.slack import MockSlackWebhook, mock_slack_webhook


@pytest.fixture
def mock_gcs() -> Iterator[MockStorageClient]:
    yield from patch_google_storage(
        expected_expiration=timedelta(hours=1), bucket_name="some-bucket"
    )


@pytest.fixture
def mock_kubernetes() -> Iterator[MockKubernetesApi]:
    yield from patch_kubernetes()


@pytest.fixture
def mock_slack(respx_mock: respx.Router) -> MockSlackWebhook:
    return mock_slack_webhook("https://example.com/slack", respx_mock)


@pytest_asyncio.fixture
async def redis_client() -> AsyncIterator[redis.Redis]:
    """Redis client for testing.

    This fixture connects to the Redis server that runs via tox-docker.
    """
    client = redis.Redis(host="localhost", port=6379, db=0)
    yield client

    await client.aclose()



"""Safir is the Rubin Observatory's library for building FastAPI services
for the Rubin Science Platform.
"""

__all__ = ["__version__", "version_info"]

from importlib.metadata import PackageNotFoundError, version

__version__: str
"""The version string of Safir (PEP 440 / SemVer compatible)."""

try:
    __version__ = version(__name__)
except PackageNotFoundError:
    # package is not installed
    __version__ = "0.0.0"

version_info = __version__.split(".")
"""The decomposed version, split across "``.``."

Use this for version comparison.
"""



"""Test Slack client."""

from __future__ import annotations

from unittest.mock import ANY

import pytest
import respx
import structlog
from httpx import AsyncClient, HTTPError, Response
from pydantic import ValidationError

from safir.slack.blockkit import (
    SlackBaseField,
    SlackCodeBlock,
    SlackCodeField,
    SlackException,
    SlackMessage,
    SlackTextBlock,
    SlackTextField,
    SlackWebException,
)
from safir.slack.webhook import SlackWebhookClient
from safir.testing.slack import MockSlackWebhook


def test_message() -> None:
    message = SlackMessage(
        message="This is some *Slack message*  \n  ",
        fields=[
            SlackTextField(heading="Some text", text="Value of the field   "),
            SlackCodeField(heading="Some code", code="Here is\nthe code\n"),
        ],
        blocks=[SlackTextBlock(heading="Log", text="Some\nlong\nlog")],
        attachments=[
            SlackCodeBlock(heading="Backtrace", code="Some\nbacktrace"),
            SlackTextBlock(heading="Essay", text="Blah blah blah"),
        ],
    )
    assert message.to_slack() == {
        "blocks": [
            {
                "type": "section",
                "text": {
                    "type": "mrkdwn",
                    "text": "This is some *Slack message*",
                    "verbatim": True,
                },
            },
            {
                "type": "section",
                "fields": [
                    {
                        "type": "mrkdwn",
                        "text": "*Some text*\nValue of the field",
                        "verbatim": True,
                    },
                    {
                        "type": "mrkdwn",
                        "text": "*Some code*\n```\nHere is\nthe code\n```",
                        "verbatim": True,
                    },
                ],
            },
            {
                "type": "section",
                "text": {
                    "type": "mrkdwn",
                    "text": "*Log*\nSome\nlong\nlog",
                    "verbatim": True,
                },
            },
        ],
        "attachments": [
            {
                "blocks": [
                    {
                        "type": "section",
                        "text": {
                            "type": "mrkdwn",
                            "text": "*Backtrace*\n```\nSome\nbacktrace\n```",
                            "verbatim": True,
                        },
                    },
                    {
                        "type": "section",
                        "text": {
                            "type": "mrkdwn",
                            "text": "*Essay*\nBlah blah blah",
                            "verbatim": True,
                        },
                    },
                ]
            }
        ],
    }

    message = SlackMessage(message="Single line message", verbatim=False)
    assert message.to_slack() == {
        "blocks": [
            {
                "type": "section",
                "text": {
                    "type": "mrkdwn",
                    "text": "Single line message",
                    "verbatim": False,
                },
            }
        ]
    }

    message = SlackMessage(
        message="Message with <special> & one `field`",
        fields=[SlackTextField(heading="Something", text="Blah")],
    )
    assert message.to_slack() == {
        "blocks": [
            {
                "type": "section",
                "text": {
                    "type": "mrkdwn",
                    "text": "Message with &lt;special&gt; &amp; one `field`",
                    "verbatim": True,
                },
            },
            {
                "type": "section",
                "fields": [
                    {
                        "type": "mrkdwn",
                        "text": "*Something*\nBlah",
                        "verbatim": True,
                    },
                ],
            },
            {"type": "divider"},
        ]
    }

    message = SlackMessage(
        message="Message with one block",
        blocks=[SlackTextBlock(heading="Something", text="Blah")],
    )
    assert message.to_slack() == {
        "blocks": [
            {
                "type": "section",
                "text": {
                    "type": "mrkdwn",
                    "text": "Message with one block",
                    "verbatim": True,
                },
            },
            {
                "type": "section",
                "text": {
                    "type": "mrkdwn",
                    "text": "*Something*\nBlah",
                    "verbatim": True,
                },
            },
            {"type": "divider"},
        ]
    }


def test_validation() -> None:
    """Test errors caught by validation."""
    fields: list[SlackBaseField] = [
        SlackTextField(heading="Something", text="foo")
    ] * 11
    message = SlackMessage(message="Ten fields", fields=fields[:10])
    assert len(message.fields) == 10
    with pytest.raises(ValidationError):
        SlackMessage(message="Eleven fields", fields=fields)


def test_block_truncation() -> None:
    """Test truncating attachments at Slack limits."""
    block = SlackTextBlock(heading="Something", text="a" * 3000)
    length = 3000 - len("*Something*\n\n... truncated ...")
    assert block.to_slack()["text"] == (
        "*Something*\n" + "a" * length + "\n... truncated ..."
    )

    block = SlackTextBlock(heading="Something", text="abcde\n" * 500)
    length = int((3001 - len("*Something*\n\n... truncated ...")) / 6)
    assert block.to_slack()["text"] == (
        "*Something*\n" + "abcde\n" * length + "... truncated ..."
    )

    cblock = SlackCodeBlock(heading="Else", code="a" * 3000)
    length = 3000 - len("*Else*\n```\n... truncated ...\n\n```")
    assert cblock.to_slack()["text"] == (
        "*Else*\n```\n... truncated ...\n" + "a" * length + "\n```"
    )

    cblock = SlackCodeBlock(heading="Else", code="abcde\n" * 500)
    length = int((3001 - len("*Else*\n```\n... truncated ...\n\n```")) / 6)
    assert cblock.to_slack()["text"] == (
        "*Else*\n```\n... truncated ...\n"
        + ("abcde\n" * length).strip()
        + "\n```"
    )


def test_field_truncation() -> None:
    """Test truncating fields at Slack limits."""
    field = SlackTextField(heading="Something", text="a" * 2000)
    length = 2000 - len("*Something*\n\n... truncated ...")
    assert field.to_slack()["text"] == (
        "*Something*\n" + "a" * length + "\n... truncated ..."
    )

    field = SlackTextField(heading="Something", text="abcdefg\n" * 250)
    length = int((2001 - len("*Something*\n\n... truncated ...")) / 8)
    assert field.to_slack()["text"] == (
        "*Something*\n" + "abcdefg\n" * length + "... truncated ..."
    )

    cfield = SlackCodeField(heading="Else", code="a" * 2000)
    length = 2000 - len("*Else*\n```\n... truncated ...\n\n```")
    assert cfield.to_slack()["text"] == (
        "*Else*\n```\n... truncated ...\n" + "a" * length + "\n```"
    )

    cfield = SlackCodeField(heading="Else", code="abcdefg\n" * 250)
    length = int((2001 - len("*Else*\n```\n... truncated ...\n\n```")) / 8)
    assert cfield.to_slack()["text"] == (
        "*Else*\n```\n... truncated ...\n"
        + ("abcdefg\n" * length).strip()
        + "\n```"
    )


def test_message_truncation() -> None:
    """Text truncating the main part of a Slack message."""
    message = SlackMessage(message="a" * 3000)
    assert message.to_slack()["blocks"][0]["text"]["text"] == "a" * 3000
    message = SlackMessage(message="a" * 3001)
    length = 3000 - len("\n... truncated ...")
    assert message.to_slack()["blocks"][0]["text"]["text"] == (
        "a" * length + "\n... truncated ..."
    )


@pytest.mark.asyncio
async def test_exception(mock_slack: MockSlackWebhook) -> None:
    logger = structlog.get_logger(__file__)
    slack = SlackWebhookClient(mock_slack.url, "App", logger)

    class SomeError(SlackException):
        pass

    try:
        raise SomeError("Some error", "someuser")
    except SlackException as e:
        await slack.post_exception(e)

    assert mock_slack.messages == [
        {
            "blocks": [
                {
                    "type": "section",
                    "text": {
                        "type": "mrkdwn",
                        "text": "Error in App: Some error",
                        "verbatim": True,
                    },
                },
                {
                    "type": "section",
                    "fields": [
                        {
                            "type": "mrkdwn",
                            "text": "*Exception type*\nSomeError",
                            "verbatim": True,
                        },
                        {
                            "type": "mrkdwn",
                            "text": ANY,
                            "verbatim": True,
                        },
                        {
                            "type": "mrkdwn",
                            "text": "*User*\nsomeuser",
                            "verbatim": True,
                        },
                    ],
                },
                {"type": "divider"},
            ],
        }
    ]


@pytest.mark.asyncio
async def test_web_exception(
    respx_mock: respx.Router, mock_slack: MockSlackWebhook
) -> None:
    logger = structlog.get_logger(__file__)
    slack = SlackWebhookClient(mock_slack.url, "App", logger)

    class SomeError(SlackWebException):
        pass

    respx_mock.get("https://example.org/").mock(return_value=Response(404))
    exception = None
    try:
        async with AsyncClient() as client:
            r = await client.get("https://example.org/")
            r.raise_for_status()
    except HTTPError as e:
        exception = SomeError.from_exception(e)
        assert str(exception) == "Status 404 from GET https://example.org/"
        await slack.post_exception(exception)

    assert mock_slack.messages == [
        {
            "blocks": [
                {
                    "type": "section",
                    "text": {
                        "type": "mrkdwn",
                        "text": "Error in App: " + str(exception),
                        "verbatim": True,
                    },
                },
                {
                    "type": "section",
                    "fields": [
                        {
                            "type": "mrkdwn",
                            "text": "*Exception type*\nSomeError",
                            "verbatim": True,
                        },
                        {
                            "type": "mrkdwn",
                            "text": ANY,
                            "verbatim": True,
                        },
                    ],
                },
                {
                    "type": "section",
                    "text": {
                        "type": "mrkdwn",
                        "text": "*URL*\nGET https://example.org/",
                        "verbatim": True,
                    },
                },
                {"type": "divider"},
            ],
        }
    ]



"""Utility functions for Pydantic models."""

from __future__ import annotations

from collections.abc import Callable
from datetime import UTC, datetime
from typing import Any, ParamSpec, TypeVar

from pydantic import BaseModel, ConfigDict

P = ParamSpec("P")
T = TypeVar("T")

__all__ = [
    "CamelCaseModel",
    "normalize_datetime",
    "normalize_isodatetime",
    "to_camel_case",
    "validate_exactly_one_of",
]


def normalize_datetime(v: Any) -> datetime | None:
    """Pydantic field validator for datetime fields.

    Supports `~datetime.datetime` fields given as either datetime objects or
    seconds since epoch (not the other types Pydantic natively supports) and
    ensures that the resulting datetime object is timezone-aware and in the
    UTC timezone.

    Parameters
    ----------
    v
        Field representing a `~datetime.datetime`.

    Returns
    -------
    datetime.datetime or None
        The timezone-aware `~datetime.datetime` or `None` if the input was
        `None`.

    Examples
    --------
    Here is a partial model that uses this function as a field validator.

    .. code-block:: python

       class Info(BaseModel):
           last_used: datetime | None = Field(
               None,
               title="Last used",
               description="When last used in seconds since epoch",
               examples=[1614986130],
           )

           _normalize_last_used = field_validator("last_used", mode="before")(
               normalize_datetime
           )
    """
    if v is None:
        return v
    elif isinstance(v, int):
        return datetime.fromtimestamp(v, tz=UTC)
    elif not isinstance(v, datetime):
        raise ValueError("Must be a datetime or seconds since epoch")
    elif v.tzinfo and v.tzinfo.utcoffset(v) is not None:
        return v.astimezone(UTC)
    else:
        return v.replace(tzinfo=UTC)


def normalize_isodatetime(v: str | None) -> datetime | None:
    """Pydantic field validator for datetime fields in ISO format.

    This field validator requires the ISO 8601 date and time format with ``Z``
    as the time zone (``YYYY-MM-DDTHH:MM:SSZ``). This format is compatible
    with Kubernetes and the ISO UWS standard and is the same format produced
    by `safir.datetime.isodatetime`. It should be used when the ambiguous
    formats supported by Pydantic by default (such as dates and times without
    time zone information) shouldn't be allowed.

    Parameters
    ----------
    v
        The field representing a `~datetime.datetime`.

    Returns
    -------
    datetime.datetime or None
        The timezone-aware `~datetime.datetime` or `None` if the input was
        `None`.

    Examples
    --------
    Here is a partial model that uses this function as a field validator.

    .. code-block:: python

       class Info(BaseModel):
           last_used: datetime | None = Field(
               None,
               title="Last used",
               description="Date and time last used",
               examples=["2023-01-25T15:44:34Z"],
           )

           _normalize_last_used = field_validator("last_used", mode="before")(
               normalize_isodatetime
           )
    """
    if v is None:
        return None
    if not isinstance(v, str) or not v.endswith("Z"):
        raise ValueError("Must be a string in YYYY-MM-DDTHH:MM[:SS]Z format")
    try:
        return datetime.fromisoformat(v[:-1] + "+00:00")
    except Exception as e:
        raise ValueError(f"Invalid date {v}: {e!s}") from e


def to_camel_case(string: str) -> str:
    """Convert a string to camel case.

    Intended for use with Pydantic as an alias generator so that the model can
    be initialized from camel-case input, such as Kubernetes objects or
    settings from Helm charts.

    Parameters
    ----------
    string
        Input string.

    Returns
    -------
    str
        String converted to camel-case with the first character in lowercase.

    Examples
    --------
    To support ``camelCase`` input to a model, use the following settings:

    .. code-block:: python

       class Model(BaseModel):
           some_field: str

           model_config = ConfigDict(
               alias_generator=to_camel_case, populate_by_name=True
           )

    This must be added to every class that uses ``snake_case`` for an
    attribute and that needs to be initialized from ``camelCase``.
    Alternately, inherit from `~safir.pydantic.CamelCaseModel`, which is
    derived from `pydantic.BaseModel` with those settings added.
    """
    components = string.split("_")
    return components[0] + "".join(c.title() for c in components[1:])


def _copy_type(
    parent: Callable[P, T],
) -> Callable[[Callable[..., T]], Callable[P, T]]:
    """Copy the type of a parent method.

    Used to avoid duplicating the prototype of `pydantic.BaseModel.dict` and
    `pydantic.BaseModel.json` when overriding them in `CamelCaseModel`.

    Parameters
    ----------
    parent
        Method from which to copy a type signature.

    Returns
    -------
    Callable
        Decorator that will replace the type signature of a function with the
        type signature of its parent.
    """
    return lambda x: x


class CamelCaseModel(BaseModel):
    """`pydantic.BaseModel` configured to accept camel-case input.

    This is a convenience class identical to `~pydantic.BaseModel` except with
    an alias generator configured so that it can be initialized with either
    camel-case or snake-case keys. Model exports with ``model_dump`` or
    ``model_dump_json`` also default to exporting in camel-case.
    """

    model_config = ConfigDict(
        alias_generator=to_camel_case, populate_by_name=True
    )

    @_copy_type(BaseModel.model_dump)
    def model_dump(self, **kwargs: Any) -> dict[str, Any]:
        """Export the model as a dictionary.

        Overridden to change the default of ``by_alias`` from `False` to
        `True`, so that by default the exported dictionary uses camel-case.
        """
        if "by_alias" not in kwargs:
            kwargs["by_alias"] = True
        return super().model_dump(**kwargs)

    @_copy_type(BaseModel.model_dump_json)
    def model_dump_json(self, **kwargs: Any) -> str:
        """Export the model as JSON.

        Overridden to change the default of ``by_alias`` from `False` to
        `True`, so that by default the exported dictionary uses camel-case.
        """
        if "by_alias" not in kwargs:
            kwargs["by_alias"] = True
        return super().model_dump_json(**kwargs)


def validate_exactly_one_of(
    *settings: str,
) -> Callable[[BaseModel], BaseModel]:
    """Generate a model validator imposing a one and only one constraint.

    Sometimes, models have a set of attributes of which one and only one may
    be set. Ideally this is represented properly in the type system, but
    occasionally it's more convenient to use a model validator. This is a
    model validator generator that can produce a model validator function that
    ensures one and only one of an arbitrary set of attributes must be set.

    Parameters
    ----------
    *settings
        List of names of attributes, of which one and only one must be set.
        At least two attribute names must be listed.

    Returns
    -------
    Callable
        Resulting model validator.

    Examples
    --------
    Use this inside a Pydantic class as a model validator as follows:

    .. code-block:: python

       class Foo(BaseModel):
           foo: Optional[str] = None
           bar: Optional[str] = None
           baz: Optional[str] = None

           _validate_options = model_validator(mode="after")(
               validate_exactly_one_of("foo", "bar", "baz")
           )

    The attribute listed as the first argument to the ``model_validator`` call
    must be the last attribute in the model definition so that any other
    attributes have already been seen.
    """
    if len(settings) < 2:
        msg = "validate_exactly_one_of takes at least two field names"
        raise ValueError(msg)

    if len(settings) == 2:
        options = f"{settings[0]} and {settings[1]}"
    else:
        options = ", ".join(settings[:-1]) + ", and " + settings[-1]

    def validator(model: T) -> T:
        seen = False
        for setting in settings:
            if getattr(model, setting, None) is not None:
                if seen:
                    raise ValueError(f"only one of {options} may be given")
                seen = True
        if not seen:
            raise ValueError(f"one of {options} must be given")
        return model

    return validator



"""An arq_ client with a mock for testing."""

from __future__ import annotations

import abc
import uuid
from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from typing import Any, Self

from arq import create_pool
from arq.connections import ArqRedis, RedisSettings
from arq.constants import default_queue_name as arq_default_queue_name
from arq.jobs import Job, JobStatus

from .datetime import current_datetime

__all__ = [
    "ArqJobError",
    "JobNotQueued",
    "JobNotFound",
    "JobResultUnavailable",
    "ArqMode",
    "JobMetadata",
    "JobResult",
    "ArqQueue",
    "RedisArqQueue",
    "MockArqQueue",
]


class ArqJobError(Exception):
    """A base class for errors related to arq jobs.

    Attributes
    ----------
    job_id
        The job ID, or `None` if the job ID is not known in this context.
    """

    def __init__(self, message: str, job_id: str | None) -> None:
        super().__init__(message)
        self._job_id = job_id

    @property
    def job_id(self) -> str | None:
        """The job ID, or `None` if the job ID is not known in this context."""
        return self._job_id


class JobNotQueued(ArqJobError):
    """The job was not successfully queued."""

    def __init__(self, job_id: str | None) -> None:
        super().__init__(
            f"Job was not queued because it already exists. id={job_id}",
            job_id,
        )


class JobNotFound(ArqJobError):
    """A job cannot be found."""

    def __init__(self, job_id: str) -> None:
        super().__init__(f"Job could not be found. id={job_id}", job_id)


class JobResultUnavailable(ArqJobError):
    """The job's result is unavailable."""

    def __init__(self, job_id: str) -> None:
        super().__init__(f"Job result could not be found. id={job_id}", job_id)


class ArqMode(str, Enum):
    """Mode configuration for the Arq queue."""

    production = "production"
    """Normal usage of arq, with a Redis broker."""

    test = "test"
    """Use the MockArqQueue to test an API service without standing up a
    full distributed worker queue.
    """


@dataclass
class JobMetadata:
    """Information about a queued job.

    Attributes
    ----------
    id
        The `arq.jobs.Job` identifier
    name
        The task name.
    args
        The positional arguments to the task function.
    kwargs
        The keyword arguments to the task function.
    enqueue_time
        Time when the job was added to the queue.
    status
        Status of the job.

        States are defined by the `arq.jobs.JobStatus` enumeration:

        - ``deferred`` (in queue, but waiting a predetermined time to become
          ready to run)
        - ``queued`` (queued to run)
        - ``in_progress`` (actively being run by a worker)
        - ``complete`` (result is available)
        - ``not_found`` (the job cannot be found)
    queue_name
        Name of the queue this job belongs to.
    """

    id: str
    """The `~arq.jobs.Job` identifier."""

    name: str
    """The task name."""

    args: tuple[Any, ...]
    """The positional arguments to the task function."""

    kwargs: dict[str, Any]
    """The keyword arguments to the task function."""

    enqueue_time: datetime
    """Time when the job was added to the queue."""

    status: JobStatus
    """Status of the job.

    States are defined by the `arq.jobs.JobStatus` enumeration:

    - ``deferred`` (in queue, but waiting a predetermined time to become
      ready to run)
    - ``queued`` (queued to run)
    - ``in_progress`` (actively being run by a worker)
    - ``complete`` (result is available)
    - ``not_found`` (the job cannot be found)
    """

    queue_name: str
    """Name of the queue this job belongs to."""

    @classmethod
    async def from_job(cls, job: Job) -> Self:
        """Initialize JobMetadata from an arq Job.

        Raises
        ------
        JobNotFound
            Raised if the job is not found
        """
        job_info = await job.info()
        if job_info is None:
            raise JobNotFound(job.job_id)

        job_status = await job.status()
        if job_status == JobStatus.not_found:
            raise JobNotFound(job.job_id)

        return cls(
            id=job.job_id,
            name=job_info.function,
            args=job_info.args,
            kwargs=job_info.kwargs,
            enqueue_time=job_info.enqueue_time,
            status=job_status,
            # private attribute of Job; not available in JobDef
            # queue_name is available in JobResult
            queue_name=job._queue_name,  # noqa: SLF001
        )


@dataclass
class JobResult(JobMetadata):
    """The full result of a job, as well as its metadata.

    Attributes
    ----------
    id
        The `~arq.jobs.Job` identifier
    name
        The task name.
    args
        The positional arguments to the task function.
    kwargs
        The keyword arguments to the task function.
    enqueue_time
        Time when the job was added to the queue.
    status
        Status of the job.

        States are defined by the `arq.jobs.JobStatus` enumeration:

        - ``deferred`` (in queue, but waiting a predetermined time to become
          ready to run)
        - ``queued`` (queued to run)
        - ``in_progress`` (actively being run by a worker)
        - ``complete`` (result is available)
        - ``not_found`` (the job cannot be found)
    start_time
        Time when the job started.
    finish_time
        Time when the job finished.
    success
        `True` if the job returned without an exception, `False` if an
        exception was raised.
    result
        The job's result.
    """

    start_time: datetime
    """Time when the job started."""

    finish_time: datetime
    """Time when the job finished."""

    success: bool
    """`True` if the job returned without an exception, `False` if an
    exception was raised.
    """

    result: Any
    """The job's result."""

    @classmethod
    async def from_job(cls, job: Job) -> Self:
        """Initialize the `JobResult` from an arq `~arq.jobs.Job`.

        Raises
        ------
        JobNotFound
            Raised if the job is not found
        JobResultUnavailable
            Raised if the job result is not available.
        """
        job_info = await job.info()
        if job_info is None:
            raise JobNotFound(job.job_id)

        job_status = await job.status()
        if job_status == JobStatus.not_found:
            raise JobNotFound(job.job_id)

        # Result may be none if the job isn't finished
        result_info = await job.result_info()
        if result_info is None:
            raise JobResultUnavailable(job.job_id)

        return cls(
            id=job.job_id,
            name=job_info.function,
            args=job_info.args,
            kwargs=job_info.kwargs,
            enqueue_time=job_info.enqueue_time,
            start_time=result_info.start_time,
            finish_time=result_info.finish_time,
            success=result_info.success,
            status=job_status,
            queue_name=result_info.queue_name,
            result=result_info.result,
        )


class ArqQueue(metaclass=abc.ABCMeta):
    """A common interface for working with an arq queue that can be
    implemented either with a real Redis backend, or an in-memory repository
    for testing.

    See Also
    --------
    RedisArqQueue
        Production implementation with a Redis store.
    MockArqQueue
        In-memory implementation for testing and development.
    """

    def __init__(
        self, *, default_queue_name: str = arq_default_queue_name
    ) -> None:
        self._default_queue_name = default_queue_name

    @property
    def default_queue_name(self) -> str:
        """Name of the default queue, if the ``_queue_name`` parameter is not
        set in method calls.
        """
        return self._default_queue_name

    @abc.abstractmethod
    async def enqueue(
        self,
        task_name: str,
        *task_args: Any,
        _queue_name: str | None = None,
        **task_kwargs: Any,
    ) -> JobMetadata:
        """Add a job to the queue.

        Parameters
        ----------
        task_name
            The function name to run.
        *args
            Positional arguments for the task function.
        _queue_name
            Name of the queue.
        **kwargs
            Keyword arguments passed to the task function.

        Returns
        -------
        JobMetadata
            Metadata about the queued job.

        Raises
        ------
        JobNotQueued
            Raised if the job is not successfully added to the queue.
        """
        raise NotImplementedError

    @abc.abstractmethod
    async def get_job_metadata(
        self, job_id: str, queue_name: str | None = None
    ) -> JobMetadata:
        """Get metadata about a `~arq.jobs.Job`.

        Parameters
        ----------
        job_id
            The job's identifier. This is the same as the `JobMetadata.id`
            attribute, provided when initially adding a job to the queue.
        queue_name
            Name of the queue.

        Returns
        -------
        JobMetadata
            Metadata about the queued job.

        Raises
        ------
        JobNotFound
            Raised if the job is not found in the queue.
        """
        raise NotImplementedError

    @abc.abstractmethod
    async def get_job_result(
        self, job_id: str, queue_name: str | None = None
    ) -> JobResult:
        """Retrieve the job result, if available.

        Parameters
        ----------
        job_id
            The job's identifier. This is the same as the `JobMetadata.id`
            attribute, provided when initially adding a job to the queue.
        queue_name
            Name of the queue.

        Returns
        -------
        JobResult
            The job's result, along with metadata about the queued job.

        Raises
        ------
        JobNotFound
            Raised if the job is not found in the queue.
        JobResultUnavailable
            Raised if the job's result is unavailable for any reason.
        """
        raise NotImplementedError


class RedisArqQueue(ArqQueue):
    """A distributed queue, based on arq and Redis."""

    def __init__(
        self,
        pool: ArqRedis,
        *,
        default_queue_name: str = arq_default_queue_name,
    ) -> None:
        super().__init__(default_queue_name=default_queue_name)
        self._pool = pool

    @classmethod
    async def initialize(
        cls,
        redis_settings: RedisSettings,
        *,
        default_queue_name: str = arq_default_queue_name,
    ) -> Self:
        """Initialize a RedisArqQueue from Redis settings."""
        pool = await create_pool(
            redis_settings, default_queue_name=default_queue_name
        )
        return cls(pool)

    async def enqueue(
        self,
        task_name: str,
        *task_args: Any,
        _queue_name: str | None = None,
        **task_kwargs: Any,
    ) -> JobMetadata:
        job = await self._pool.enqueue_job(
            task_name,
            *task_args,
            _queue_name=_queue_name or self.default_queue_name,
            **task_kwargs,
        )
        if job:
            return await JobMetadata.from_job(job)
        else:
            # TODO(jonathansick): if implementing hard-coded job IDs, set as
            # an argument
            raise JobNotQueued(None)

    def _get_job(self, job_id: str, queue_name: str | None = None) -> Job:
        return Job(
            job_id,
            self._pool,
            _queue_name=queue_name or self.default_queue_name,
        )

    async def get_job_metadata(
        self, job_id: str, queue_name: str | None = None
    ) -> JobMetadata:
        job = self._get_job(job_id, queue_name=queue_name)
        return await JobMetadata.from_job(job)

    async def get_job_result(
        self, job_id: str, queue_name: str | None = None
    ) -> JobResult:
        job = self._get_job(job_id, queue_name=queue_name)
        return await JobResult.from_job(job)


class MockArqQueue(ArqQueue):
    """A mocked queue for testing API services."""

    def __init__(
        self, *, default_queue_name: str = arq_default_queue_name
    ) -> None:
        super().__init__(default_queue_name=default_queue_name)
        self._job_metadata: dict[str, dict[str, JobMetadata]] = {
            self.default_queue_name: {}
        }
        self._job_results: dict[str, dict[str, JobResult]] = {
            self.default_queue_name: {}
        }

    def _resolve_queue_name(self, queue_name: str | None) -> str:
        return queue_name or self.default_queue_name

    async def enqueue(
        self,
        task_name: str,
        *task_args: Any,
        _queue_name: str | None = None,
        **task_kwargs: Any,
    ) -> JobMetadata:
        queue_name = self._resolve_queue_name(_queue_name)
        new_job = JobMetadata(
            id=str(uuid.uuid4().hex),
            name=task_name,
            args=task_args,
            kwargs=task_kwargs,
            enqueue_time=current_datetime(microseconds=True),
            status=JobStatus.queued,
            queue_name=queue_name,
        )
        self._job_metadata[queue_name][new_job.id] = new_job
        return new_job

    async def get_job_metadata(
        self, job_id: str, queue_name: str | None = None
    ) -> JobMetadata:
        queue_name = self._resolve_queue_name(queue_name)
        try:
            return self._job_metadata[queue_name][job_id]
        except KeyError as e:
            raise JobNotFound(job_id) from e

    async def get_job_result(
        self, job_id: str, queue_name: str | None = None
    ) -> JobResult:
        queue_name = self._resolve_queue_name(queue_name)
        try:
            return self._job_results[queue_name][job_id]
        except KeyError as e:
            raise JobResultUnavailable(job_id) from e

    async def set_in_progress(
        self, job_id: str, queue_name: str | None = None
    ) -> None:
        """Set a job's status to in progress, for mocking a queue in tests."""
        job = await self.get_job_metadata(job_id, queue_name=queue_name)
        job.status = JobStatus.in_progress

        # An in-progress job cannot have a result
        if job_id in self._job_results:
            del self._job_results[job_id]

    async def set_complete(
        self,
        job_id: str,
        *,
        result: Any,
        success: bool = True,
        queue_name: str | None = None,
    ) -> None:
        """Set a job's result, for mocking a queue in tests."""
        queue_name = self._resolve_queue_name(queue_name)

        job_metadata = await self.get_job_metadata(
            job_id, queue_name=queue_name
        )
        job_metadata.status = JobStatus.complete

        result_info = JobResult(
            id=job_metadata.id,
            name=job_metadata.name,
            args=job_metadata.args,
            kwargs=job_metadata.kwargs,
            status=job_metadata.status,
            enqueue_time=job_metadata.enqueue_time,
            start_time=current_datetime(microseconds=True),
            finish_time=current_datetime(microseconds=True),
            result=result,
            success=success,
            queue_name=queue_name,
        )
        self._job_results[queue_name][job_id] = result_info



"""Test the Gafaelfawr auth FastAPI dependencies."""

from __future__ import annotations

import json
from typing import Annotated
from unittest.mock import ANY

import pytest
from _pytest.logging import LogCaptureFixture
from fastapi import Depends, FastAPI
from httpx import ASGITransport, AsyncClient
from structlog.stdlib import BoundLogger

from safir.dependencies.gafaelfawr import (
    auth_delegated_token_dependency,
    auth_dependency,
    auth_logger_dependency,
)
from safir.logging import configure_logging


@pytest.mark.asyncio
async def test_auth_dependency() -> None:
    app = FastAPI()

    @app.get("/")
    async def handler(
        user: Annotated[str, Depends(auth_dependency)],
    ) -> dict[str, str]:
        return {"user": user}

    transport = ASGITransport(app=app)  # type: ignore[arg-type]
    base_url = "https://example.com"
    async with AsyncClient(transport=transport, base_url=base_url) as client:
        r = await client.get("/")
        assert r.status_code == 422

        r = await client.get("/", headers={"X-Auth-Request-User": "someuser"})
        assert r.status_code == 200
        assert r.json() == {"user": "someuser"}


@pytest.mark.asyncio
async def test_auth_delegated_token_dependency() -> None:
    app = FastAPI()

    @app.get("/")
    async def handler(
        token: Annotated[str, Depends(auth_delegated_token_dependency)],
    ) -> dict[str, str]:
        return {"token": token}

    transport = ASGITransport(app=app)  # type: ignore[arg-type]
    base_url = "https://example.com"
    async with AsyncClient(transport=transport, base_url=base_url) as client:
        r = await client.get("/")
        assert r.status_code == 422

        r = await client.get(
            "/", headers={"X-Auth-Request-Token": "sometoken"}
        )
        assert r.status_code == 200
        assert r.json() == {"token": "sometoken"}


@pytest.mark.asyncio
async def test_auth_logger_dependency(caplog: LogCaptureFixture) -> None:
    configure_logging(name="myapp", profile="production", log_level="info")

    app = FastAPI()

    @app.get("/")
    async def handler(
        logger: Annotated[BoundLogger, Depends(auth_logger_dependency)],
    ) -> dict[str, str]:
        logger.info("something")
        return {}

    caplog.clear()
    transport = ASGITransport(app=app)  # type: ignore[arg-type]
    base_url = "https://example.com"
    async with AsyncClient(transport=transport, base_url=base_url) as client:
        r = await client.get("/", headers={"User-Agent": ""})
        assert r.status_code == 422

        r = await client.get(
            "/", headers={"User-Agent": "", "X-Auth-Request-User": "someuser"}
        )
        assert r.status_code == 200

    assert len(caplog.record_tuples) == 1
    assert json.loads(caplog.record_tuples[0][2]) == {
        "event": "something",
        "httpRequest": {
            "requestMethod": "GET",
            "requestUrl": "https://example.com/",
            "remoteIp": "127.0.0.1",
        },
        "logger": "myapp",
        "request_id": ANY,
        "severity": "info",
        "user": "someuser",
    }



"""Test IVOA middleware."""

from __future__ import annotations

from typing import Annotated

import pytest
from fastapi import FastAPI, Query
from httpx import ASGITransport, AsyncClient

from safir.middleware.ivoa import CaseInsensitiveQueryMiddleware


def build_app() -> FastAPI:
    """Construct a test FastAPI app with the middleware registered."""
    app = FastAPI()
    app.add_middleware(CaseInsensitiveQueryMiddleware)
    return app


@pytest.mark.asyncio
async def test_case_insensitive() -> None:
    app = build_app()

    @app.get("/")
    async def handler(param: str) -> dict[str, str]:
        return {"param": param}

    @app.get("/simple")
    async def simple_handler() -> dict[str, str]:
        return {"foo": "bar"}

    @app.get("/list")
    async def list_handler(
        param: Annotated[list[str], Query()],
    ) -> dict[str, list[str]]:
        return {"param": param}

    transport = ASGITransport(app=app)  # type: ignore[arg-type]
    base_url = "https://example.com"
    async with AsyncClient(transport=transport, base_url=base_url) as client:
        r = await client.get("/", params={"param": "foo"})
        assert r.status_code == 200
        assert r.json() == {"param": "foo"}

        r = await client.get("/", params={"PARAM": "foo"})
        assert r.status_code == 200
        assert r.json() == {"param": "foo"}

        r = await client.get("/", params={"pARam": "foo"})
        assert r.status_code == 200
        assert r.json() == {"param": "foo"}

        r = await client.get("/", params={"paramX": "foo"})
        assert r.status_code == 422

        r = await client.get("/simple")
        assert r.status_code == 200
        assert r.json() == {"foo": "bar"}

        r = await client.get(
            "/list",
            params=[("param", "foo"), ("PARAM", "BAR"), ("parAM", "baZ")],
        )
        assert r.status_code == 200
        assert r.json() == {"param": ["foo", "BAR", "baZ"]}



"""Utilities for configuring structlog-based logging."""

from __future__ import annotations

import logging
import logging.config
import re
import sys
from enum import Enum
from typing import Any, Self

import structlog
from structlog.stdlib import add_log_level
from structlog.types import EventDict

__all__ = [
    "LogLevel",
    "Profile",
    "add_log_severity",
    "configure_logging",
    "configure_uvicorn_logging",
    "logger_name",
]

logger_name: str | None = None
"""Name of the configured global logger.

When `configure_logging` is called, the name of the configured logger is
stored in this variable. It is used by the logger dependency to retrieve the
application's configured logger.

Only one configured logger is supported. Additional calls to
`configure_logging` change the stored logger name.
"""

_UVICORN_ACCESS_REGEX = re.compile(r'^[0-9.]+:[0-9]+ - "([^"]+)" ([0-9]+)$')
"""Regex to parse Uvicorn access logs."""


class Profile(Enum):
    """Logging profile for the application."""

    production = "production"
    """Log messages in JSON."""

    development = "development"
    """Log messages in a format intended for human readability."""


class LogLevel(Enum):
    """Python logging level.

    Any case variation is accepted when converting a string to an enum value
    via the class constructor.
    """

    DEBUG = "DEBUG"
    INFO = "INFO"
    WARNING = "WARNING"
    ERROR = "ERROR"
    CRITICAL = "CRITICAL"

    @classmethod
    def _missing_(cls, value: Any) -> Self | None:
        """Allow strings in any case to be used to create the enum."""
        if not isinstance(value, str):
            return None
        value = value.upper()
        for member in cls:
            if member.value == value:
                return member
        return None


def add_log_severity(
    logger: logging.Logger, method_name: str, event_dict: EventDict
) -> EventDict:
    """Add the log level to the event dict as ``severity``.

    Intended for use as a structlog processor.

    This is the same as `structlog.stdlib.add_log_level` except that it
    uses the ``severity`` key rather than ``level`` for compatibility with
    Google Log Explorer and its automatic processing of structured logs.

    Parameters
    ----------
    logger
        The wrapped logger object.
    method_name
        The name of the wrapped method (``warning`` or ``error``, for
        example).
    event_dict
        Current context and current event. This parameter is also modified in
        place, matching the normal behavior of structlog processors.

    Returns
    -------
    ``structlog.types.EventDict``
        The modified ``structlog.types.EventDict`` with the added key.
    """
    severity = add_log_level(logger, method_name, {})["level"]
    event_dict["severity"] = severity
    return event_dict


def configure_logging(
    *,
    name: str,
    profile: Profile | str = Profile.production,
    log_level: LogLevel | str = LogLevel.INFO,
    add_timestamp: bool = False,
) -> None:
    """Configure logging and structlog.

    Parameters
    ----------
    name
        Name of the logger, which is typically the name of your application's
        root namespace.
    profile
        The name of the application profile:

        development
            Log messages are formatted for easier reading on the terminal.
        production
            Log messages are formatted as JSON objects.

        May be given as a `Profile` enum value (preferred) or a string.
    log_level
        The Python log level. May be given as a `LogLevel` enum (preferred)
        or a case-insensitive string.
    add_timestamp
        Whether to add an ISO-format timestamp to each log message.

    Notes
    -----
    This function helps you configure a useful logging set up for your
    application that's based on structlog_.

    First, it configures the logger for your application to log to STDOUT.
    Second, it configures the formatting of your log messages through
    structlog.

    In development mode, messages are key-value formatted, like this:

    .. code-block:: text

       [info     ] Hello world                    [myapp] answer=42

    Here, "Hello world" is the message. ``answer=42`` is a value bound to the
    logger.

    In production mode, messages are formatted as JSON objects:

    .. code-block:: text

       {"answer": 42, "event": "Hello world", "logger": "myapp",
       "severity": "info"}

    Examples
    --------
    .. code-block:: python

       import structlog
       from safir.logging import configure_logging


       configure_logging(name="mybot")
       logger = structlog.get_logger("mybot")
       logger.info("Hello world")
    """
    if not isinstance(log_level, LogLevel):
        log_level = LogLevel[log_level.upper()]
    if not isinstance(profile, Profile):
        profile = Profile[profile]

    stream_handler = logging.StreamHandler(stream=sys.stdout)
    stream_handler.setFormatter(logging.Formatter("%(message)s"))
    logger = logging.getLogger(name)
    logger.handlers = []
    logger.addHandler(stream_handler)
    logger.setLevel(log_level.value)

    processors: list[Any] = [
        structlog.stdlib.filter_by_level,
        structlog.stdlib.add_logger_name,
        structlog.stdlib.PositionalArgumentsFormatter(),
    ]
    if add_timestamp:
        processors.append(structlog.processors.TimeStamper(fmt="iso"))
    processors.extend(
        [
            structlog.processors.StackInfoRenderer(),
            structlog.processors.UnicodeDecoder(),
        ]
    )
    if profile == Profile.production:
        # JSON-formatted logging
        processors.append(add_log_severity)
        processors.append(structlog.processors.format_exc_info)
        processors.append(structlog.processors.JSONRenderer())
    else:
        # Key-value formatted logging
        processors.append(structlog.stdlib.add_log_level)
        processors.append(structlog.dev.ConsoleRenderer())

    structlog.configure(
        processors=processors,
        logger_factory=structlog.stdlib.LoggerFactory(),
        wrapper_class=structlog.stdlib.BoundLogger,
        cache_logger_on_first_use=True,
    )

    # Set the configured name for the global logger.
    global logger_name
    logger_name = name


def _process_uvicorn_access_log(
    logger: logging.Logger, method_name: str, event_dict: EventDict
) -> EventDict:
    """Parse a Uvicorn access log entry into key/value pairs.

    Intended for use as a structlog processor.

    This checks whether the log message is a Uvicorn access log entry and, if
    so, parses the message into key/value pairs for JSON logging so that the
    details can be programmatically extracted.  ``remoteIp`` is intentionally
    omitted since it isn't aware of ``X-Forwarded-For`` and will therefore
    always point to an uninteresting in-cluster IP.

    Parameters
    ----------
    logger
        The wrapped logger object.
    method_name
        The name of the wrapped method (``warning`` or ``error``, for
        example).
    event_dict
        Current context and current event. This parameter is also modified in
        place, matching the normal behavior of structlog processors.

    Returns
    -------
    EventDict
        The modified `~structlog.types.EventDict` with the added key.
    """
    match = _UVICORN_ACCESS_REGEX.match(event_dict["event"])
    if not match:
        return event_dict
    request = match.group(1)
    method, rest = request.split(" ", 1)
    url, protocol = rest.rsplit(" ", 1)
    if "httpRequest" not in event_dict:
        event_dict["httpRequest"] = {}
    event_dict["httpRequest"]["protocol"] = protocol
    event_dict["httpRequest"]["requestMethod"] = method
    event_dict["httpRequest"]["requestUrl"] = url
    event_dict["httpRequest"]["status"] = match.group(2)
    return event_dict


def configure_uvicorn_logging(
    log_level: LogLevel | str = LogLevel.INFO,
) -> None:
    """Set up logging.

    This configures Uvicorn to use structlog for output formatting and
    installs a custom processor to parse its access log messages into
    additional log context that matches the format of Google log messages.
    This helps Google's Cloud Logging system understand the logs. Access logs
    are sent to standard output and all other logs are sent to standard
    error.

    Parameters
    ----------
    log_level
        The Python log level. May be given as a `LogLevel` enum (preferred)
        or a case-insensitive string.

    Notes
    -----
    This method should normally be called during FastAPI app creation, either
    during Python module import or inside the function that creates and
    returns the FastAPI app that Uvicorn will run. This ensures the logging
    setup is complete before Uvicorn logs its first message.
    """
    if not isinstance(log_level, LogLevel):
        log_level = LogLevel[log_level.upper()]

    processors = [
        structlog.stdlib.ProcessorFormatter.remove_processors_meta,
        structlog.processors.format_exc_info,
        structlog.processors.JSONRenderer(),
    ]
    logging.config.dictConfig(
        {
            "version": 1,
            "disable_existing_loggers": False,
            "formatters": {
                "json-access": {
                    "()": structlog.stdlib.ProcessorFormatter,
                    "processors": processors,
                    "foreign_pre_chain": [
                        add_log_severity,
                        _process_uvicorn_access_log,
                    ],
                },
                "json": {
                    "()": structlog.stdlib.ProcessorFormatter,
                    "processors": processors,
                    "foreign_pre_chain": [add_log_severity],
                },
            },
            "handlers": {
                "uvicorn.access": {
                    "level": log_level.value,
                    "class": "logging.StreamHandler",
                    "formatter": "json-access",
                    "stream": "ext://sys.stdout",
                },
                "uvicorn.default": {
                    "level": log_level.value,
                    "class": "logging.StreamHandler",
                    "formatter": "json",
                },
            },
            "loggers": {
                "uvicorn.error": {
                    "handlers": ["uvicorn.default"],
                    "level": log_level.value,
                    "propagate": False,
                },
                "uvicorn.access": {
                    "handlers": ["uvicorn.access"],
                    "level": log_level.value,
                    "propagate": False,
                },
            },
        }
    )
